/*
  This header file defines "portable" socket library, which
  can be used without any changes on Windows and UNIX operating
  systems.
  
  This header file should be included instead of ordinary
  Winsock2.h in Windows and all necessary header files in UNIX
  (such as <sys/types.h>, <sys/socket.h>, <netinet/in.h> and so on).
  
  After including this header file you can use native socket API
  (socket(), bind(), connect(), etc) with slight modifications noted
  below or you can use class Socket that is also defined here.

  Using native API: when using native API you should follow this rules:
  1) socket type is PR_SOCKET
  2) all socket functions (except socket() and accept()) in case of error
     return PR_SOCKET_ERROR
  3) functions socket() and accept() in case of error return PR_INVALID_SOCKET
  4) all socket initialization routines are called in pr_init()
     function and thus performed during PR library initialization phase
*/

#ifndef _PR_SOCKET_H
#define _PR_SOCKET_H

#include <errno.h>
#include "pr_general.h"

// defining all necessary macros for Windows and UNIX operating systems
#ifdef WIN32
// macros and headers for Windows platform
#include <winsock.h>
#define	PR_SOCKET SOCKET
#define	PR_SOCKET_ERROR SOCKET_ERROR
#define	PR_INVALID_SOCKET INVALID_SOCKET
#define close closesocket
#else
// macros and headers for UNIX platform
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#define PR_SOCKET int
#define PR_SOCKET_ERROR -1
#define PR_INVALID_SOCKET -1
#endif // WIN32


//  Abstract class Socket. Supports only Internet TCP sockets
class Socket: public pr{
protected:
	PR_SOCKET sock_fd;		// socket descriptor
	struct sockaddr_in local;	// sockaddr_in structure for local socket
	struct sockaddr_in peer;	// sockaddr_in structure for peer
	bool created;			// state flag

	// objects of type Socket can not be directly instantiated
	Socket();
	~Socket();
public:
	// make all calls to the socket blocking
	bool makeBlocking(void);
	// make all calls to the socket nonblocking
	bool makeNonBlocking(void);
	// returns true if last error was caused by nonblocking socket
	bool wouldBlock(void);
	// close connection
	bool Close(void);
	// return true if socket is created succesfully
	bool Created(void);
};

// Standard socket for read-write operations
class stdSocket: public Socket{
public:
	stdSocket(): Socket(){}
	stdSocket(PR_SOCKET sock, struct sockaddr_in* local, struct sockaddr_in* peer);
	~stdSocket(){}
	bool Connect(char* hostname, int Port);
	int Send(char* Mes, int Length);
	int Recv(char* Buf, int Lenght);
	char* GetRemoteIp(void);
};

// Accept socket only for creating new sockets
class acceptSocket: public Socket{
public:
	acceptSocket(int Port);
	~acceptSocket(){}
	stdSocket* Accept(void);
private:
	acceptSocket(){}
};

#endif // _PR_SOCKET_H
