#include "pr_server.h"
#include "pr_general.h"
#include <iostream>
#include <string>
#include <string.h>

using namespace std;

Server::Server(const char* serverName, int _argc, char** _argv){
	ServerName = serverName;
	argc = _argc;
	argv = _argv;
#ifdef WIN32
	__server = this;
#endif // WIN32
}

const char* Server::GetServerName(){
	return ServerName.c_str();
}

bool Server::Start(){
#ifdef WIN32
	SERVICE_STATUS ssStatus;
	LPQUERY_SERVICE_CONFIG lpqscBuf;
    DWORD dwBytesNeeded;
	char pathToService[256];
	
	GetFullPath(pathToService, argv[0], sizeof(pathToService));
	if (pathToService[0] == '\0'){
		cerr << "Can not start service. Can not get full path to service." << endl;
		return false;
	}
	strcat(pathToService, " start");

	// Look params
	if (argc > 1){
		// If there is something and this something equals to "start" that
		// it means that SCM tries to start us - so let's be polite and do it
		if (strncmp(argv[1], "start", 5) != 0){
			// error
			// Something strange has happened. Won't risk and return false
			return false;
		}
		// and so it begins
		WinService();
	} else {
		// If nothing than we should tell SCM to start the service
		
		// first of all we should open SCM database and look for our service
		SC_HANDLE schSCManager = OpenSCManager(
			NULL,					// Local machine
			NULL,					// ServiceACtive database
			SC_MANAGER_ALL_ACCESS);	// full access rights
		if (schSCManager == NULL){
			// error
			cerr << "Can not start service. Can not open SCM database" << endl;
			return false;
		}

		SC_HANDLE schService = OpenService(
			schSCManager,	// Service Control Manager
			(LPCTSTR)ServerName.c_str(),	// Service name
			SERVICE_QUERY_CONFIG |		// Access rights
			SERVICE_CHANGE_CONFIG |
			SERVICE_QUERY_STATUS |
			SERVICE_START);

		if (schService == NULL){
			// some error occured. Let's find out what it is
			if (GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST){
				// this is the first time we start our service - so we should
				// register it in SCM database
				schService = CreateService(
					schSCManager,	// Service Control Manager
					(LPCTSTR)ServerName.c_str(),	// Service name
					(LPCTSTR)ServerName.c_str(),	// Service name that would be displayed by Windows
					SERVICE_ALL_ACCESS,		// Access rights
					SERVICE_WIN32_OWN_PROCESS,	// Service type
					SERVICE_DEMAND_START,	// Should be launched manually
					SERVICE_ERROR_NORMAL,	// Error control type
					(LPCTSTR)pathToService,
					NULL,	// No load ordering group
					NULL,	// No tag identifiers
					NULL,	// No dependencies
					NULL,	// LocalSystem account
					NULL);	// No password
				if (schService == NULL){
					// error
					cerr << "Can not start service. Can not register it in SCM database." << endl;
					CloseServiceHandle(schSCManager);
					return false;
				}
				if (ServiceStart(schService) == NO_ERROR){
					CloseServiceHandle(schSCManager);
					CloseServiceHandle(schService);
					cerr << "Service started" << endl;
					return true;
				} else {
					cerr << "Can not start service." << endl;
					CloseServiceHandle(schSCManager);
					CloseServiceHandle(schService);
					return false;
				}
			} else {
				// something fatal has happened. So we shoud stop
				CloseServiceHandle(schSCManager);
				cerr << "Can not start service. Error code: " << GetLastError() << endl;
				return false;
			}
		} else {
			// We have opened successfully our service, so let's continue our work
			if (!QueryServiceStatus(schService, &ssStatus)){
				// error
				cerr << "Can not start service. Can not query service status." << endl;
				CloseServiceHandle(schService);
				CloseServiceHandle(schSCManager);
				return false;
			}
			
			if (ssStatus.dwCurrentState == SERVICE_RUNNING){
				// Service is already running. Let's tell this great news to the user and go home
				cerr << "The service \'" << ServerName.c_str() << "\' has already been started" << endl;
				CloseServiceHandle(schService);
				CloseServiceHandle(schSCManager);
				return false;
			}

			lpqscBuf = (LPQUERY_SERVICE_CONFIG) new char[4096];
			memset((void*)lpqscBuf, 0, 4096);
			if (lpqscBuf == NULL){
				// error
				cerr << "Can not start service. Can not allocate enough memory." << endl;
				CloseServiceHandle(schService);
				CloseServiceHandle(schSCManager);
				return false;
			}
			
			// Get configuration information
			if (!QueryServiceConfig(schService,	lpqscBuf, 4096, &dwBytesNeeded)){
				// error
				cerr << "Can not start service. Can not query service configuration." << endl;
				CloseServiceHandle(schService);
				CloseServiceHandle(schSCManager);
				return false;
			}

			char answer;
			if (strcmp(lpqscBuf->lpBinaryPathName, pathToService) != 0){
				// Service is executed from different path as it was earlier
				// It is either because we are trying to executing something not concerned with our
				// service, or because it was moved to another location.
				// In any case we should ask the user what to do
				cerr << "Path to service or starting argument have changed. There are three possible reasons:" << endl;
				cerr << "1) You are trying to start the service with the name that already belongs to another service" << endl;
				cerr << "2) Binary file was moved to another location" << endl;
				cerr << "3) Starting argument has changed" << endl << endl;
				cerr << "Previous path: " << lpqscBuf->lpBinaryPathName << endl;
				cerr << "Current path: " << pathToService << endl << endl;
				cerr << "Register new path and start the service? [y/n] ";
				cin >> answer;
				if (answer == 'y'){
					// Set new path
					if (ServiceSetNewPath(pathToService, schSCManager, schService) != NO_ERROR){
						// error
						cerr << "Can not start service. Can not set new path." << endl;
						return false;
					}
				} else {
					// error
					cerr << "Service not started." << endl;
					CloseServiceHandle(schService);
					CloseServiceHandle(schSCManager);
					return false;
				}
			}

			// Start the service and exit
			if (ServiceStart(schService) != NO_ERROR){
				// error
				cerr << "Can not start service." << endl;
				CloseServiceHandle(schService);
				CloseServiceHandle(schSCManager);
				return false;
			}
			CloseServiceHandle(schService);
			CloseServiceHandle(schSCManager);
			cerr << "Service started" << endl;
			return true;
		}
	}
	return true;
#else
	int pid = fork();
	if (pid == 0)
		Run();
	return true;
#endif // WIN32
}

/////////////////////////////////////////////////////////////////
// Defining Windows-specific functions that work with services //
/////////////////////////////////////////////////////////////////

#ifdef WIN32
void Server::WinService(){
	SERVICE_TABLE_ENTRY DispatchTable[] = {{(LPTSTR)ServerName.c_str(), (LPSERVICE_MAIN_FUNCTIONA)ServiceMain}, {NULL, NULL}};

	if (!StartServiceCtrlDispatcher(DispatchTable)){
		cerr << "StartServiceCtrlDispatcher error: can not start service" << endl;
		exit(0);
	}
}

void ServiceMain(DWORD argc, LPTSTR* argv){
	MyServiceStatus.dwServiceType = SERVICE_WIN32;
	MyServiceStatus.dwCurrentState = SERVICE_START_PENDING;
	MyServiceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	MyServiceStatus.dwWin32ExitCode = 0;
	MyServiceStatus.dwServiceSpecificExitCode = 0;
	MyServiceStatus.dwCheckPoint = 0;
	MyServiceStatus.dwWaitHint = 0;

	if (__server == NULL){
		// error
		return;
	}

	MyServiceStatusHandle = RegisterServiceCtrlHandler(
		__server->GetServerName(),
		(LPHANDLER_FUNCTION)ServiceCtrlHandler);

	if (MyServiceStatusHandle == (SERVICE_STATUS_HANDLE)0)
		return;

	MyServiceStatus.dwCurrentState = SERVICE_RUNNING;
	MyServiceStatus.dwCheckPoint = 0;
	MyServiceStatus.dwWaitHint = 0;

	if (!SetServiceStatus(MyServiceStatusHandle, &MyServiceStatus))
		return;
	
	__server->Run();

	return;
}

void ServiceCtrlHandler(DWORD Opcode){
	switch(Opcode){
		case SERVICE_CONTROL_STOP:
			MyServiceStatus.dwWin32ExitCode = 0;
			MyServiceStatus.dwCurrentState = SERVICE_STOPPED;
			MyServiceStatus.dwCheckPoint = 0;
			MyServiceStatus.dwWaitHint = 0;
			
			__server->OnTerminate();
			
			if (!SetServiceStatus(MyServiceStatusHandle, &MyServiceStatus)){
				// error
				return;
			}
		case SERVICE_CONTROL_INTERROGATE:
			__server->OnReload();
			break;
	}
	
	if (!SetServiceStatus(MyServiceStatusHandle, &MyServiceStatus)){
		// error
	}
	return;
}

DWORD ServiceStart(SC_HANDLE schService){
	SERVICE_STATUS ssStatus; 
    DWORD dwOldCheckPoint; 
    DWORD dwStartTickCount;
    DWORD dwWaitTime;

	if (!StartService(
			schService,	// Handle to service
			0,	// Number of arguments
			NULL))	// No arguments
	{
		// error
		return GetLastError();
	}
	
	// Service start pending
	// Check the status until the service is no longer start pending
	if (!QueryServiceStatus(
			schService,	// Handle to service
			&ssStatus))	// Address of status information structure
	{
		// error
		return GetLastError();
	}

	// Save the tickpoint and initial checkpoint
	dwStartTickCount = GetTickCount();
	dwOldCheckPoint = ssStatus.dwCheckPoint;

	while (ssStatus.dwCurrentState == SERVICE_START_PENDING){
		// Do not wait longer than the wait hint. A good interval is
		// one tenth the wait hint, but no less that 1 second and
		// no more than 10 seconds.
		dwWaitTime = ssStatus.dwWaitHint / 10;

		if (dwWaitTime < 1000)
			dwWaitTime = 1000;
		else if (dwWaitTime > 10000)
			dwWaitTime = 10000;

		Sleep(dwWaitTime);

		// Check the status again

		if (!QueryServiceStatus(
				schService,	// Handle to service
				&ssStatus))	// Addres of structure
			break;

		if (ssStatus.dwCheckPoint > dwOldCheckPoint){
			// The service is making progress
			dwStartTickCount = GetTickCount();
			dwOldCheckPoint = ssStatus.dwCheckPoint;
		}
		else {
			if (GetTickCount() - dwStartTickCount > ssStatus.dwWaitHint){
				// No progress made within the wait hint
				break;
			}
		}
	}

	if (ssStatus.dwCurrentState == SERVICE_RUNNING)
		return NO_ERROR;	// everything is OK

	return GetLastError();
}

DWORD ServiceSetNewPath(const char* path, SC_HANDLE schSCManager, SC_HANDLE schService){
	SC_LOCK sclLock; 
	DWORD error;

	// Need to lock database before continue
	sclLock = LockServiceDatabase(schSCManager);

	if (sclLock == NULL){
		error = GetLastError();
		int time = 0;
		while (error == ERROR_SERVICE_DATABASE_LOCKED){
			time++;
			Sleep(1000);
			sclLock = LockServiceDatabase(schSCManager);
			error = GetLastError();
			if (sclLock != NULL)
				break;
			if (time > 10){
				// database is locked for more then 10 seconds. It's abnormal so we should terminate
				return error;
			}
		}
		if (sclLock == NULL){
			// error
			return error;
		}
	}

	// Lock obtained. We can continue change configuration
	if (!ChangeServiceConfig(
			schService,	// Handle to service
			SERVICE_NO_CHANGE,	// No change
			SERVICE_NO_CHANGE,	// No change
			SERVICE_NO_CHANGE,	// No change
			(LPCTSTR)path,		// New path
			NULL,	// No change
			NULL,	// No change
			NULL,	// No change
			NULL,	// No change
			NULL,	// No change
			NULL))	// No change
	{
		error = GetLastError();
		UnlockServiceDatabase(sclLock);
		return error;
	}
	UnlockServiceDatabase(sclLock);
	return NO_ERROR;
}
#endif // WIN32